# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2019 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
"""
Responsabilities

    Handles accumulation of debug events while playing media_player and  saves
    when sample's play ends.
"""
from __future__ import division, print_function, unicode_literals

import pickle
import time


class BufferedLogger(object):
    def __init__(self, outfile):
        self.outfile = outfile
        self.log_entries = []
        self.start_wall_time = None
        # (fn, args)
        self.on_close_callback_info = None

    def init_wall_time(self):
        self.start_wall_time = time.perf_counter()

    def log(self, *args):
        self.log_entries.append(args)

    def rebased_wall_time(self):
        return time.perf_counter() - self.start_wall_time

    def close(self):
        self.save_log_entries_as_pickle()
        if self.on_close_callback_info is not None:
            fn, args = self.on_close_callback_info
            fn(self.log_entries, *args)

    def save_log_entries_as_pickle(self):
        with open(self.outfile, "wb") as f:
            pickle.dump(self.log_entries, f)

    def clear(self):
        self.log_entries = []


logger = None



# -*- coding: utf-8 -*-
''' pyglet specific docstring transformations.
'''

_debug = False

def debug(lines):
    with open('debug.log', 'a') as f:
        for line in lines:
            f.write(line+"\n")

if _debug:
    with open('debug.log', 'w') as f:
        f.write("Docstring modifications.\n\n")


def indentation(line):
    if line.strip()=="": return 0
    return len(line) - len(line.lstrip())


def process_block(converter, lines, start):
    ''' Apply a transformation to an indented block
    '''
    first = indentation(lines[start])
    current = first
    blocks = [[]]
    block = 0

    for i, line in enumerate(lines[start+1:]):
        level = indentation(line)
        if level<=first:
            try: # allow one blank line in the block
                if line.strip()=="" and \
                   (current == indentation(lines[start + i + 2])):
                    continue
            except: pass
            break
        if level<current:
            blocks.append([line])
            block += 1
        else:
            blocks[block].append(line)
        current = level

    result = []
    for block in blocks:
            result += converter(block)

    return i, result


def ReST_parameter(lines):
    ''' Converts :parameters: blocks to :param: markup
    '''
    indent = indentation(lines[0])
    part = lines[0].replace("`","").split(":")
    name = part[0].replace(" ", "")
    rest = [":param "+name+":"]
    for line in lines[1:]:
        rest.append(line[indent:])
    if len(part)>1:
        rest.append(":type "+name+": "+part[1].strip().lstrip())
    return rest


def ReST_Ivariable(lines):
    ''' Converts :Ivariable: blocks to :var: markup
    '''
    indent = indentation(lines[0])
    part = lines[0].replace("`","").split(":")
    name = part[0].replace(" ", "")
    rest = [":var "+name+":"]
    for line in lines[1:]:
        rest.append(line[indent:])
    if len(part)>1:
        rest.append(":vartype "+name+": "+part[1].strip().lstrip())
    return rest




def modify_docstrings(app, what, name, obj, options, lines,
                      reference_offset=[0]):

    original = lines[:]

    def convert(converter, start):
        affected, result = process_block(converter, lines, start)
        for x in range(start, start+affected+1):
            del lines[start]
        for i, line in enumerate(result):
            lines.insert(start+i, line)

    i=0
    while i<len(lines):
        line = lines[i]
        if ":parameters:" in line.lower():
            convert(ReST_parameter, i)

        elif ":Ivariables:" in line:
            convert(ReST_Ivariable, i)

        elif ":deprecated:" in line:
            lines[i] = lines[i].replace(u':deprecated:',
                                u'.. warning:: Deprecated.')
            lines.insert(i,"")
            lines.insert(i,"")

        elif ":event:" in line.lower():
            lines[i] = lines[i].replace(u':event:', u'.. event mark')

        i += 1


    if _debug and original!=lines:
        title = what + " " +name
        debug(["\n",title, "-"*len(title)])
        debug(["Original:", ""]+original)
        debug(["Redacted:", ""]+lines)


def setup(app):
    app.connect('autodoc-process-docstring', modify_docstrings)



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2019 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

'''Compatibility tools

Various tools for simultaneous Python 2.x and Python 3.x support
'''

__docformat__ = 'restructuredtext'
__version__ = '$Id$'

import sys
import itertools
import weakref

if sys.version_info[0] == 2:
    if sys.version_info[1] < 6:
        # Pure Python implementation from
        # http://docs.python.org/library/itertools.html#itertools.izip_longest
        def izip_longest(*args, **kwds):
            # izip_longest('ABCD', 'xy', fillvalue='-') --> Ax By C- D-
            fillvalue = kwds.get('fillvalue')
            def sentinel(counter=([fillvalue] * (len(args) - 1)).pop):
                yield counter()     # yields the fillvalue, or raises IndexError
            fillers = itertools.repeat(fillvalue)
            iters = [itertools.chain(it, sentinel(), fillers) for it in args]
            try:
                for tup in itertools.izip(*iters):
                    yield tup
            except IndexError:
                pass
    else:
        izip_longest = itertools.izip_longest
else:
    izip_longest = itertools.zip_longest


if sys.version_info[0] >= 3:
    import io

    def asbytes(s):
        if isinstance(s, bytes):
            return s
        elif isinstance(s, str):
            return bytes(ord(c) for c in s)
        else:
            return bytes(s)

    def asbytes_filename(s):
        if isinstance(s, bytes):
            return s
        elif isinstance(s, str):
            return s.encode(encoding=sys.getfilesystemencoding())

    def asstr(s):
        if s is None:
            return ''
        if isinstance(s, str):
            return s
        return s.decode("utf-8")

    bytes_type = bytes
    BytesIO = io.BytesIO
else:
    import StringIO

    asbytes = str
    asbytes_filename = str
    asstr = str
    bytes_type = str
    BytesIO = StringIO.StringIO


# Backporting for Python < 3.4
class _WeakMethod(weakref.ref):
    """
    A custom 'weakref.ref' subclass which simulates a weak reference to
    a bound method, working around the lifetime problem of bound methods
    """

    __slots__ = '_func_ref', '_meth_type', '_alive', '__weakref__'

    def __new__(cls, meth, callback=None):
        try:
            obj = meth.__self__
            func = meth.__func__
        except AttributeError:
            raise TypeError('argument should be a bound method, not {0}'.format(type(meth)))

        def _cb(arg):
            # The self-weakref trick is needed to avoid creating a reference cycle.
            self = self_wr()
            if self._alive:
                self._alive = False
                if callback is not None:
                    callback(self)
        self = weakref.ref.__new__(cls, obj, _cb)
        self._func_ref = weakref.ref(func, _cb)
        self._meth_type = type(meth)
        self._alive = True
        self_wr = weakref.ref(self)
        return self

    def __call__(self):
        obj = super(WeakMethod, self).__call__()
        func = self._func_ref()
        if obj is None or func is None:
            return None
        return self._meth_type(func, obj)

    def __eq__(self, other):
        if isinstance(other, WeakMethod):
            if not self._alive or not other._alive:
                return self is other
            return weakref.ref.__eq__(self, other) and self._func_ref == other._func_ref
        return False

    def __ne__(self, other):
        if isinstance(other, WeakMethod):
            if not self._alive or not other._alive:
                return self is not other
            return weakref.ref.__ne__(self, other) or self._func_ref != other._func_ref
        return True

    __hash__ = weakref.ref.__hash__


if sys.version_info < (3, 4):
    WeakMethod = _WeakMethod
else:
    from weakref import WeakMethod



# Copyright 2009 Joe Wreschnig and others.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The names of the authors and contributors may be used to endorse or
#    promote products derived from this software without specific prior
#    written permission.
#
# All warranties and liabilities, express or implied, are hereby
# disclaimed.

"""Draw a NinePatch image.

NinePatch is a format for storing how to cut up a 9-part resizable
rectangular image within the image pixel data directly.

For more information on the NinePatch format, see
http://developer.android.com/guide/topics/graphics/2d-graphics.html#nine-patch.

"""

__all__ = ["NinePatch"]

from pyglet.gl import *

class PixelData(object):
    def __init__(self, image):
        image_data = image.get_image_data()
        self.has_alpha = 'A' in image_data.format
        self.data = image_data.get_data("RGBA", image.width * 4)
        self.width = image.width
        self.height = image.height

    def is_black(self, x, y):
        p = (y * self.width + x) * 4
        if self.has_alpha:
            if self.data[p+3] == '\x00':
                return False # Fully transparent

        return self.data[p:p+3] == '\x00\x00\x00'

class NinePatch(object):
    """A scalable 9-patch image.
    """

    # Content area of the image, in pixels from the edge.
    padding_top = None
    padding_bottom = None
    padding_right = None
    padding_left = None

    # Resizable area of the image, in pixels from the closest edge
    stretch_top = None
    stretch_left = None
    stretch_right = None
    stretch_bottom = None

    def __init__(self, image):
        """Create NinePatch cuts of an image

        Arguments:
            image - an ImageData (Texture, TextureRegion, etc)
            texture - force cut ImageDatas to be Textures (or Regions)
        """

        data = PixelData(image)
        width = data.width
        height = data.height

        # Texture dimensions after removing the 9patch outline.
        self.width = width - 2
        self.height = height - 2

        # Only need to retain the texture for drawing
        self.texture = image.get_texture()

        # Find stretch area markers
        for x in range(1, width - 1):
            if data.is_black(x, height - 1):
                self.stretch_left = x
                break
        else:
            self.stretch_left = 1

        for x in range(width - 2, 0, -1):
            if data.is_black(x, height - 1):
                self.stretch_right = width - x
                break
        else:
            self.stretch_right = 1

        for y in range(1, height - 1):
            if data.is_black(0, y):
                self.stretch_bottom = y
                break
        else:
            self.stretch_bottom = 1

        for y in range(height - 2, 0, -1):
            if data.is_black(0, y):
                self.stretch_top = height - y
                break
        else:
            self.stretch_top = 1

        # Find content area markers, if any
        for x in range(1, width - 1):
            if data.is_black(x, 0):
                self.padding_left = x - 1
                break

        for x in range(width - 2, 0, -1):
            if data.is_black(x, 0):
                self.padding_right = self.width - x
                break

        for y in range(1, height - 1):
            if data.is_black(width - 1, y):
                self.padding_bottom = y - 1
                break

        for y in range(height - 2, 0, -1):
            if data.is_black(width - 1, y):
                self.padding_top = self.height - y
                break

        # Texture coordinates, in pixels
        u1 = 1
        v1 = 1
        u2 = self.stretch_left + 1
        v2 = self.stretch_bottom + 1
        u3 = width - self.stretch_right - 1
        v3 = height - self.stretch_top - 1
        u4 = width - 1
        v4 = height - 1

        # Texture coordinates as ratio of image size (0 to 1)
        u1, u2, u3, u4 = [s / float(width) for s in (u1, u2, u3, u4)]
        v1, v2, v3, v4 = [s / float(height) for s in (v1, v2, v3, v4)]

        # Scale texture coordinates to match the tex_coords pyglet gives us
        # (these aren't necessarily 0-1 as the texture may have been packed)
        (tu1, tv1, _, 
         _, _, _, 
         tu2, tv2, _, 
         _, _, _) = self.texture.tex_coords
        u_scale = tu2 - tu1
        u_bias = tu1
        v_scale = tv2 - tv1
        v_bias = tv1
        u1, u2, u3, u4 = [u_bias + u_scale * s for s in (u1, u2, u3, u4)]
        v1, v2, v3, v4 = [v_bias + v_scale * s for s in (v1, v2, v3, v4)]

        # 2D texture coordinates, bottom-left to top-right
        self.tex_coords = (
            u1, v1,
            u2, v1,
            u3, v1,
            u4, v1,
            u1, v2,
            u2, v2,
            u3, v2,
            u4, v2,
            u1, v3,
            u2, v3,
            u3, v3,
            u4, v3,
            u1, v4,
            u2, v4,
            u3, v4,
            u4, v4,
        )

        # Quad indices
        self.indices = []
        for y in range(3):
            for x in range(3):
                self.indices.extend([
                    x + y * 4,
                    (x + 1) + y * 4,
                    (x + 1) + (y + 1) * 4,
                    x + (y + 1) * 4,
                ])

    def get_vertices(self, x, y, width, height):
        """Get 16 2D vertices for the given image region"""
        x1 = x
        y1 = y
        x2 = x + self.stretch_left
        y2 = y + self.stretch_bottom
        x3 = x + width - self.stretch_right
        y3 = y + height - self.stretch_top
        x4 = x + width
        y4 = y + height

        # To match tex coords, vertices are bottom-left to top-right
        return (
            x1, y1,
            x2, y1,
            x3, y1,
            x4, y1,
            x1, y2,
            x2, y2,
            x3, y2,
            x4, y2,
            x1, y3,
            x2, y3,
            x3, y3,
            x4, y3,
            x1, y4,
            x2, y4,
            x3, y4,
            x4, y4,
        )

    def draw(self, x, y, width, height):
        """Draw the nine-patch at the given image dimensions."""
        width = max(width, self.width + 2)
        height = max(height, self.height + 2)
        vertices = self.get_vertices(x, y, width, height)

        glPushAttrib(GL_ENABLE_BIT)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(self.texture.target)
        glBindTexture(self.texture.target, self.texture.id)
        pyglet.graphics.draw_indexed(
            16, GL_QUADS, self.indices,
            ('v2i', vertices),
            ('t2f', self.tex_coords))
        glPopAttrib()

    def draw_around(self, x, y, width, height):
        """Draw the nine-patch around the given content area"""
        self.draw(x - self.padding_left,
                  y - self.padding_bottom,
                  width + self.padding_left + self.padding_right,
                  height + self.padding_bottom + self.padding_top)

if __name__ == '__main__':
    import sys
    image = pyglet.image.load(sys.argv[1])
    ninepatch = NinePatch(image)

    window = pyglet.window.Window(resizable=True)
    label = pyglet.text.Label('Hello, NinePatch', 
                              font_size=16,
                              anchor_y='bottom',
                              color=(0,0,0,255))

    @window.event
    def on_draw():
        glClearColor(.7, .7, .7, 1.)
        window.clear()
        width = label.content_width
        height = label.content_height
        label.x = window.width / 2 - width / 2
        label.y = window.height / 2 - height / 2
        ninepatch.draw_around(label.x, label.y, width, height)
        label.draw()
        
    pyglet.app.run()



'''
win32priority.py
A set of utilies to patch Pyglet to provide more precise event timing on Windows.

The following options are provided:

    pin_thread: Pin the main Pyglet app thread to one or more cores by way of a thread affinity mask.
        Default is True.

    pin_thread_mask: The affinity mask for pin_thread. Default is 1.

    set_switch_interval: For Python 3.x only. Sets the sys.setswitchinterval interpreter option
        to a value that ensures all threads run to completion before being switched away.
        I have not objectively confirmed this helps performance, but I've provided it anyway.
        It might be useful on slower machines.

    switch_interval_value: The timeslice value provided to set_switch_interval.
        Default is 1/65, or just under the amount of time neede to render 60 fps.

    raise_priority: Raises the process priority to High.

    patch_timer: Adjusts the main Win32 platform event loop so that it uses the
        high-resolution multimedia timer to provide proper timings for 60fps+ framerates.

    pin_process: Pins the entire Python interpreter process to the first CPU core.
        You usually don't need to do this, but I'm providing it anyway.

    patch_event_loop: Replaces the default event loop with one that is more amenable
        to an application where you're firing the main event loop by way of the
        scheduler. If you use this, you'll need to manually call window.switch_to() and
        window.flip() in your draw routine.

'''

def priority_patch(
    pin_thread = True,
    pin_thread_mask = 0x01,
    set_switch_interval = True,
    switch_interval_value = 1./65.,
    raise_priority=True,
    priority_class = 0x080,
    patch_timer=True,
    pin_process = False,
    patch_event_loop=False,
):

    import ctypes
    try:
        _winmm = ctypes.windll.winmm
    except AttributeError:
        raise NotImplementedError("This is not a Win32 compatible platform.")
    
    if set_switch_interval:
        import sys
        sys.setswitchinterval(switch_interval_value)

    time_start = _winmm.timeBeginPeriod
    time_stop = _winmm.timeEndPeriod
    _1 = ctypes.c_uint(1)

    k32 = ctypes.windll.kernel32

    if raise_priority:
        k32.SetPriorityClass(-1, priority_class)

    import pyglet
    
    if pin_thread:
        k32.SetThreadAffinityMask(pyglet.app.platform_event_loop._event_thread,pin_thread_mask)

    if pin_process:
        k32.SetProcessAffinityMask(k32.GetCurrentProcess(), 1)

    if patch_event_loop:
        def idle(self):
            self.clock.call_scheduled_functions(self.clock.update_time())
            return self.clock.get_sleep_time(True)
        
        pyglet.app.EventLoop.idle = idle

    if patch_timer:
        old_step = pyglet.clock.Clock.sleep

        def new_step(*a, **ka):
            time_start(_1)
            old_step(*a,**ka)
            time_stop(_1)

        pyglet.clock.Clock.sleep=new_step


# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2019 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
'''Wrapper for include/libavcodec/avcodec.h
'''
from __future__ import unicode_literals
from ctypes import (c_int, c_uint16, c_int32, c_int64, c_uint32, c_uint64, 
    c_uint8, c_uint, c_double, c_float, c_ubyte, c_size_t, c_char, c_char_p, 
    c_void_p, addressof, byref, cast, POINTER, CFUNCTYPE, Structure, Union, 
    create_string_buffer, memmove)

import pyglet
import pyglet.lib
from . import libavutil

avcodec = pyglet.lib.load_library(
    'avcodec',
    win32='avcodec-58',
    darwin='avcodec.58'
)

FF_INPUT_BUFFER_PADDING_SIZE = 32

class AVPacketSideData(Structure):
    _fields_ = [
        ('data', POINTER(c_uint8)),
        ('size', c_int),
        ('type', c_int)
    ]
AVBufferRef = libavutil.AVBufferRef

class AVPacket(Structure):
    _fields_ = [
        ('buf', POINTER(AVBufferRef)),
        ('pts', c_int64),
        ('dts', c_int64),
        ('data', POINTER(c_uint8)),
        ('size', c_int),
        ('stream_index', c_int),
        ('flags', c_int),
        ('side_data', POINTER(AVPacketSideData)),
        ('side_data_elems', c_int),
        ('duration', c_int64),
        ('pos', c_int64),
        ('convergence_duration', c_int64) #Deprecated
    ]
class AVCodecParserContext(Structure): pass

AVRational = libavutil.AVRational
class AVCodecParameters(Structure):
    _fields_ = [
        ('codec_type', c_int),
        ('codec_id', c_int),
        ('codec_tag', c_uint32),
        ('extradata', POINTER(c_uint8)),
        ('extradata_size', c_int),
        ('format', c_int),
        ('bit_rate', c_int64),
        ('bits_per_coded_sample', c_int),
        ('bits_per_raw_sample', c_int),
        ('profile', c_int),
        ('level', c_int),
        ('width', c_int),
        ('height', c_int),
        ('sample_aspect_ratio', AVRational),
        ('field_order', c_int),
        ('color_range', c_int),
        ('color_primaries', c_int),
        ('color_trc', c_int),
        ('color_space', c_int),
        ('chroma_location', c_int),
        ('video_delay', c_int),
        ('channel_layout', c_uint64),
        ('channels', c_int),
        ('sample_rate', c_int),
        ('block_align', c_int),
        ('frame_size', c_int),
        ('initial_padding', c_int),
        ('trailing_padding', c_int),
        ('seek_preroll', c_int),
    ]
class AVProfile(Structure):
    _fields_ = [
        ('profile', c_int),
        ('name', c_char_p),
    ]
class AVCodecDescriptor(Structure):
    _fields_ = [
        ('id', c_int),
        ('type', c_int),
        ('name', c_char_p),
        ('long_name', c_char_p),
        ('props', c_int),
        ('mime_types', c_char_p),
        ('profiles', POINTER(AVProfile))
    ]
class AVCodecInternal(Structure): pass
class AVCodec(Structure):
    _fields_ = [
        ('name', c_char_p),
        ('long_name', c_char_p),
        ('type', c_int),
        ('id', c_int),
        ('capabilities', c_int),
        ('supported_framerates', POINTER(AVRational)),
        ('pix_fmts', POINTER(c_int)),
        ('supported_samplerates', POINTER(c_int)),
        ('sample_fmts', POINTER(c_int)),
        ('channel_layouts', POINTER(c_uint64)),
        ('max_lowres', c_uint8),
        # And more...
    ]
class AVCodecContext(Structure): pass
class RcOverride(Structure): pass
class AVHWAccel(Structure): pass
AVClass = libavutil.AVClass
AVFrame = libavutil.AVFrame
AV_NUM_DATA_POINTERS = libavutil.AV_NUM_DATA_POINTERS
AVCodecContext._fields_ = [
        ('av_class', POINTER(AVClass)),
        ('log_level_offset', c_int),
        ('codec_type', c_int),
        ('codec', POINTER(AVCodec)),
        ('codec_id', c_int),
        ('codec_tag', c_uint),
        ('priv_data', c_void_p),
        ('internal', POINTER(AVCodecInternal)),
        ('opaque', c_void_p),
        ('bit_rate', c_int64),
        ('bit_rate_tolerance', c_int),
        ('global_quality', c_int),
        ('compression_level', c_int),
        ('flags', c_int),
        ('flags2', c_int),
        ('extradata', POINTER(c_uint8)),
        ('extradata_size', c_int),
        ('time_base', AVRational),
        ('ticks_per_frame', c_int),
        ('delay', c_int),
        ('width', c_int),
        ('height', c_int),
        ('coded_width', c_int),
        ('coded_height', c_int),
        ('gop_size', c_int),
        ('pix_fmt', c_int),
        ('draw_horiz_band', CFUNCTYPE(None,
            POINTER(AVCodecContext), POINTER(AVFrame),
            c_int*8, c_int, c_int, c_int)),
        ('get_format', CFUNCTYPE(c_int,
            POINTER(AVCodecContext), POINTER(c_int))),
        ('max_b_frames', c_int),
        ('b_quant_factor', c_float),
        ('b_frame_strategy', c_int), #Deprecated
        ('b_quant_offset', c_float),
        ('has_b_frames', c_int),
        ('mpeg_quant', c_int), #Deprecated
        ('i_quant_factor', c_float),
        ('i_quant_offset', c_float),
        ('lumi_masking', c_float),
        ('temporal_cplx_masking', c_float),
        ('spatial_cplx_masking', c_float),
        ('p_masking', c_float),
        ('dark_masking', c_float),
        ('slice_count', c_int),
        ('prediction_method', c_int), #Deprecated
        ('slice_offset', POINTER(c_int)),
        ('sample_aspect_ratio', AVRational),
        ('me_cmp', c_int),
        ('me_sub_cmp', c_int),
        ('mb_cmp', c_int),
        ('ildct_cmp', c_int),
        ('dia_size', c_int),
        ('last_predictor_count', c_int),
        ('pre_me', c_int), #Deprecated
        ('me_pre_cmp', c_int),
        ('pre_dia_size', c_int),
        ('me_subpel_quality', c_int),
        ('me_range', c_int),
        ('slice_flags', c_int),
        ('mb_decision', c_int),
        ('intra_matrix', POINTER(c_uint16)),
        ('inter_matrix', POINTER(c_uint16)),
        ('scenechange_threshold', c_int), #Deprecated
        ('noise_reduction', c_int), #Deprecated
        ('intra_dc_precision', c_int),
        ('skip_top', c_int),
        ('skip_bottom', c_int),
        ('mb_lmin', c_int),
        ('mb_lmax', c_int),
        ('me_penalty_compensation', c_int), #Deprecated
        ('bidir_refine', c_int),
        ('brd_scale', c_int), #Deprecated
        ('keyint_min', c_int),
        ('refs', c_int),
        ('chromaoffset', c_int), #Deprecated
        ('mv0_threshold', c_int),
        ('b_sensitivity', c_int), #Deprecated
        ('color_primaries', c_int),
        ('color_trc', c_int),
        ('colorspace', c_int),
        ('color_range', c_int),
        ('chroma_sample_location', c_int),
        ('slices', c_int),
        ('field_order', c_int),
        ('sample_rate', c_int),
        ('channels', c_int),
        ('sample_fmt', c_int),
        ('frame_size', c_int),
        ('frame_number', c_int),
        ('block_align', c_int),
        ('cutoff', c_int),
        ('channel_layout', c_uint64),
        ('request_channel_layout', c_uint64),
        ('audio_service_type', c_int),
        ('request_sample_fmt', c_int),
        ('get_buffer2', CFUNCTYPE(c_int,
            POINTER(AVCodecContext), POINTER(AVFrame), c_int)),
        ('refcounted_frames', c_int), #Deprecated
        ('qcompress', c_float),
        ('qblur', c_float),
        ('qmin', c_int),
        ('qmax', c_int),
        ('max_qdiff', c_int),
        ('rc_buffer_size', c_int),
        ('rc_override_count', c_int),
        ('rc_override', POINTER(RcOverride)),
        ('rc_max_rate', c_int64),
        ('rc_min_rate', c_int64),
        ('rc_max_available_vbv_use', c_float),
        ('rc_min_vbv_overflow_use', c_float),
        ('rc_initial_buffer_occupancy', c_int),
        ('coder_type', c_int), #Deprecated
        ('context_model', c_int), #Deprecated
        ('frame_skip_threshold', c_int), #Deprecated
        ('frame_skip_factor', c_int), #Deprecated
        ('frame_skip_exp', c_int), #Deprecated
        ('frame_skip_cmp', c_int), #Deprecated
        ('trellis', c_int),
        ('min_prediction_order', c_int), #Deprecated
        ('max_prediction_order', c_int), #Deprecated
        ('timecode_frame_start', c_int64), #Deprecated
        ('rtp_callback', CFUNCTYPE(None, #Deprecated
            POINTER(AVCodecContext), c_void_p, c_int, c_int)),
        ('rtp_payload_size', c_int), #Deprecated
        ('mv_bits', c_int), #Deprecated
        ('header_bits', c_int), #Deprecated
        ('i_tex_bits', c_int), #Deprecated
        ('p_tex_bits', c_int), #Deprecated
        ('i_count', c_int), #Deprecated
        ('p_count', c_int), #Deprecated
        ('skip_count', c_int), #Deprecated
        ('misc_bits', c_int), #Deprecated
        ('frame_bits', c_int), #Deprecated
        ('stats_out', c_char_p),
        ('stats_in', c_char_p),
        ('workaround_bugs', c_int),
        ('strict_std_compliance', c_int),
        ('error_concealment', c_int),
        ('debug', c_int),
        ('err_recognition', c_int),
        ('reordered_opaque', c_int64),
        ('hwaccel', POINTER(AVHWAccel)),
        ('hwaccel_context', c_void_p),
        ('error', c_uint64 * AV_NUM_DATA_POINTERS),
        ('dct_algo', c_int),
        ('idct_algo', c_int),
        ('bits_per_coded_sample', c_int),
        ('bits_per_raw_sample', c_int),
        ('lowres', c_int),
        ('coded_frame', POINTER(AVFrame)), #Deprecated
        ('thread_count', c_int),
        ('thread_type', c_int),
        ('active_thread_type', c_int),
        ('thread_safe_callbacks', c_int),
        ('execute', CFUNCTYPE(c_int,
            POINTER(AVCodecContext), 
            CFUNCTYPE(c_int, POINTER(AVCodecContext), c_void_p),
            c_void_p, c_int, c_int, c_int)),
        ('execute2', CFUNCTYPE(c_int,
            POINTER(AVCodecContext), 
            CFUNCTYPE(c_int, POINTER(AVCodecContext), c_void_p, c_int, c_int),
            c_void_p, c_int, c_int)),
        ('nsse_weight', c_int),
        ('profile', c_int),
        ('level', c_int),
        ('skip_loop_filter', c_int),
        ('skip_idct', c_int),
        ('skip_frame', c_int),
        ('subtitle_header', POINTER(c_uint8)),
        ('subtitle_header_size', c_int),
        ('vbv_delay', c_uint64), #Deprecated
        ('side_data_only_packets', c_int), #Deprecated
        ('initial_padding', c_int),
        ('framerate', AVRational),
        #!
        ('sw_pix_fmt', c_int),
        ('pkt_timebase', AVRational),
        ('codec_dexcriptor', AVCodecDescriptor),
        ('pts_correction_num_faulty_pts', c_int64),
        ('pts_correction_num_faulty_dts', c_int64),
        ('pts_correction_last_pts', c_int64),
        ('pts_correction_last_dts', c_int64),
        ('sub_charenc', c_char_p),
        ('sub_charenc_mode', c_int),
        ('skip_alpha', c_int),
        ('seek_preroll', c_int),
        ('debug_mv', c_int),
        ('chroma_intra_matrix', POINTER(c_uint16)),
        ('dump_separator', POINTER(c_uint8)),
        ('codec_whitelist', c_char_p),
        ('properties', c_uint),
        ('coded_side_data', POINTER(AVPacketSideData)),
        ('nb_coded_side_data', c_int),
        ('hw_frames_ctx', POINTER(AVBufferRef)),
        ('sub_text_format', c_int),
        ('trailing_padding', c_int),
        ('max_pixels', c_int64),
        ('hw_device_ctx', POINTER(AVBufferRef)),
        ('hwaccel_flags', c_int),
        ('apply_cropping', c_int),
        ('extra_hw_frames', c_int)
        
        
    ]

AV_CODEC_ID_VP8 = 139
AV_CODEC_ID_VP9 = 167

avcodec.av_packet_unref.argtypes = [POINTER(AVPacket)]
avcodec.av_packet_free.argtypes = [POINTER(POINTER(AVPacket))]
avcodec.av_packet_clone.restype = POINTER(AVPacket)
avcodec.av_packet_clone.argtypes = [POINTER(AVPacket)]
avcodec.av_packet_move_ref.argtypes = [POINTER(AVPacket), POINTER(AVPacket)]
avcodec.avcodec_find_decoder.restype = POINTER(AVCodec)
avcodec.avcodec_find_decoder.argtypes = [c_int]
AVDictionary = libavutil.AVDictionary
avcodec.avcodec_open2.restype = c_int
avcodec.avcodec_open2.argtypes = [POINTER(AVCodecContext),
            POINTER(AVCodec), 
            POINTER(POINTER(AVDictionary))]
avcodec.avcodec_free_context.argtypes = [POINTER(POINTER(AVCodecContext))]
avcodec.av_packet_alloc.restype = POINTER(AVPacket)
avcodec.av_init_packet.argtypes = [POINTER(AVPacket)]
avcodec.avcodec_decode_audio4.restype = c_int
avcodec.avcodec_decode_audio4.argtypes = [POINTER(AVCodecContext),
            POINTER(AVFrame), POINTER(c_int),
            POINTER(AVPacket)]
avcodec.avcodec_decode_video2.restype = c_int
avcodec.avcodec_decode_video2.argtypes = [POINTER(AVCodecContext),
            POINTER(AVFrame), POINTER(c_int),
            POINTER(AVPacket)]
avcodec.avcodec_flush_buffers.argtypes = [POINTER(AVCodecContext)]
avcodec.avcodec_alloc_context3.restype = POINTER(AVCodecContext)
avcodec.avcodec_alloc_context3.argtypes = [POINTER(AVCodec)]
avcodec.avcodec_free_context.argtypes = [POINTER(POINTER(AVCodecContext))]
avcodec.avcodec_parameters_to_context.restype = c_int
avcodec.avcodec_parameters_to_context.argtypes = [POINTER(AVCodecContext),
            POINTER(AVCodecParameters)]
avcodec.avcodec_get_name.restype = c_char_p
avcodec.avcodec_get_name.argtypes = [c_int]
avcodec.avcodec_find_decoder_by_name.restype = POINTER(AVCodec)
avcodec.avcodec_find_decoder_by_name.argtypes = [c_char_p]

__all__ = [
'avcodec',
'FF_INPUT_BUFFER_PADDING_SIZE',
'AVPacket',
'AVCodecContext',
'AV_CODEC_ID_VP8',
'AV_CODEC_ID_VP9',
]



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2019 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

'''Extensible attributed text format for representing pyglet formatted
documents.
'''
from builtins import chr
from builtins import map

from functools import reduce
import operator
import parser
import re
import token

import pyglet

_pattern = re.compile(r'''
    (?P<escape_hex>\{\#x(?P<escape_hex_val>[0-9a-fA-F]+)\})
  | (?P<escape_dec>\{\#(?P<escape_dec_val>[0-9]+)\})
  | (?P<escape_lbrace>\{\{)
  | (?P<escape_rbrace>\}\})
  | (?P<attr>\{
        (?P<attr_name>[^ \{\}]+)\s+
        (?P<attr_val>[^\}]+)\})
  | (?P<nl_hard1>\n(?=[ \t]))
  | (?P<nl_hard2>\{\}\n)
  | (?P<nl_soft>\n(?=\S))
  | (?P<nl_para>\n\n+)
  | (?P<text>[^\{\}\n]+)
    ''', re.VERBOSE | re.DOTALL)

class AttributedTextDecoder(pyglet.text.DocumentDecoder):
    def decode(self, text, location=None):
        self.doc = pyglet.text.document.FormattedDocument()

        self.length = 0
        self.attributes = {}
        next_trailing_space = True
        trailing_newline = True

        for m in _pattern.finditer(text):
            group = m.lastgroup
            trailing_space = True
            if group == 'text':
                t = m.group('text')
                self.append(t)
                trailing_space = t.endswith(' ')
                trailing_newline = False
            elif group == 'nl_soft':
                if not next_trailing_space:
                    self.append(' ')
                trailing_newline = False
            elif group in ('nl_hard1', 'nl_hard2'):
                self.append('\n')
                trailing_newline = True
            elif group == 'nl_para':
                self.append(m.group('nl_para')[1:]) # ignore the first \n
                trailing_newline = True
            elif group == 'attr':
                try:
                    ast = parser.expr(m.group('attr_val'))
                    if self.safe(ast):
                        val = eval(ast.compile())
                    else:
                        val = None
                except (parser.ParserError, SyntaxError):
                    val = None
                name = m.group('attr_name')
                if name[0] == '.':
                    if trailing_newline:
                        self.attributes[name[1:]] = val
                    else:
                        self.doc.set_paragraph_style(self.length, self.length, 
                                                     {name[1:]: val})
                else:
                    self.attributes[name] = val
            elif group == 'escape_dec':
                self.append(chr(int(m.group('escape_dec_val'))))
            elif group == 'escape_hex':
                self.append(chr(int(m.group('escape_hex_val'), 16)))
            elif group == 'escape_lbrace':
                self.append('{')
            elif group == 'escape_rbrace':
                self.append('}')
            next_trailing_space = trailing_space

        return self.doc

    def append(self, text):
        self.doc.insert_text(self.length, text, self.attributes)
        self.length += len(text)
        self.attributes.clear()

    _safe_names = ('True', 'False', 'None')

    def safe(self, ast):
        tree = ast.totuple()
        return self.safe_node(tree)

    def safe_node(self, node):
        if token.ISNONTERMINAL(node[0]):
            return reduce(operator.and_, map(self.safe_node, node[1:]))
        elif node[0] == token.NAME:
            return node[1] in self._safe_names
        else:
            return True



#!/usr/bin/env python
import sys
from setuptools import setup, find_packages

# Bump pyglet/__init__.py version as well.
VERSION = '1.4.6'
    
# The source dist comes with batteries included, the wheel can use pip to get the rest
is_wheel = 'bdist_wheel' in sys.argv

excluded = []
if is_wheel:
    excluded.append('extlibs.future')


def exclude_package(pkg):
    for exclude in excluded:
        if pkg.startswith(exclude):
            return True
    return False


def create_package_list(base_package):
    return ([base_package] +
            [base_package + '.' + pkg
             for pkg
             in find_packages(base_package)
             if not exclude_package(pkg)])


setup_info = dict(
    # Metadata
    name='pyglet',
    version=VERSION,
    author='Alex Holkner',
    author_email='Alex.Holkner@gmail.com',
    url='http://pyglet.readthedocs.org/en/latest/',
    download_url='http://pypi.python.org/pypi/pyglet',
    project_urls={
        'Documentation': 'https://pyglet.readthedocs.io/en/latest',
        'Source': 'https://github.com/pyglet/pyglet',
        'Tracker': 'https://github.com/pyglet/pyglet/issues',
    },
    description='Cross-platform windowing and multimedia library',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Games/Entertainment',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    # Package info
    packages=create_package_list('pyglet'),

    # Add _ prefix to the names of temporary build dirs
    options={
        'build': {'build_base': '_build'},
    },
    zip_safe=True,
)

if is_wheel:
    setup_info['install_requires'] = ['future']

setup(**setup_info)
